from typing import TypeAlias

from rainbowadn.collection.pair import *
from rainbowadn.core import *
from rainbowadn.flow.verification.core import *
from rainbowadn.nullability import *
from ._bankflow import *
from ._flowbank import *
from ._flowblock import *
from ._flowcheque import *
from ._flowstandard import *

__all__ = ('BankBlock',)

Index: TypeAlias = FlowStandard[FlowBlock[Pair[FlowCheque, FlowBank]]]


class BankBlock:
    def __init__(self, reference: NullableReference[FlowBlock[Pair[FlowCheque, FlowBank]]]):
        assert isinstance(reference, NullableReference)
        self.reference = reference

    @classmethod
    def flow(cls) -> 'BankFlow':
        return BankFlow(FlowBank.empty())

    @classmethod
    def link_factory(cls) -> RainbowFactory[Pair[FlowCheque, FlowBank]]:
        return PairFactory(FlowCheque.factory(), FlowBank.factory()).loose()

    @classmethod
    def empty(cls) -> 'BankBlock':
        return cls(
            NullableReference(
                Null(),
                FlowBlockFactory(cls.link_factory()).loose()
            )
        )

    @classmethod
    def verification(cls) -> Verification[Index]:
        return FlowBlockVerification(cls.flow().link_verification()).loose()

    async def verify(self) -> bool:
        assert_true(await self.verification().verify(await FlowBlock.outer_of(self.link_factory(), self.reference)))
        return True

    async def add(self, cheque: FlowCheque) -> 'BankBlock':
        assert isinstance(cheque, FlowCheque)
        previous_link: Nullable[Pair[FlowCheque, FlowBank]] = await FlowBlock.link_of(self.reference)
        assert isinstance(previous_link, Nullable)
        previous_bank: Nullable[FlowBank]
        if previous_link.null():
            previous_bank = Null()
        else:
            previous_bank = NotNull(await (previous_link.resolve()).element1.resolve())
        assert isinstance(previous_link, Nullable)
        bank: FlowBank = await self.flow().add(previous_bank, cheque)
        assert isinstance(bank, FlowBank)
        link: Pair[FlowCheque, FlowBank] = Pair(HashPoint.of(cheque), HashPoint.of(bank))
        block: FlowBlock[Pair[FlowCheque, FlowBank]] = await FlowBlock.add_to(self.reference, HashPoint.of(link))
        assert isinstance(block, FlowBlock)
        return BankBlock(NullableReference.off(block))
